@@ -53,7 +53,7 @@
 #define CUBE_IN_DEFAULT FALSE
 
 static char *cubeImages[] = {
-    "freedesktop.png"
+    "freedesktop"
 };
 #define N_CUBE_IMAGES (sizeof (cubeImages) / sizeof (cubeImages[0]))
 
@@ -245,8 +245,7 @@ readSvgToTexture (CompScreen   *s,
 
     CUBE_SCREEN (s);
 
-    if (!openImageFile (svgFileName, &name, NULL))
-	return FALSE;
+    name = strdup (svgFileName);
 
     svgHandle = rsvg_handle_new_from_file (name, &error);
 
