@@ -70,14 +70,14 @@ handleWindowDamageRect (CompWindow *w,
 	region.rects = &region.extents;
 	region.numRects = region.size = 1;
 
-	damageWindowRegion (w, &region);
-
-	if (initial)
-	    damageWindowOutputExtents (w);
+	damageScreenRegion (w->screen, &region);
     }
 
     if (!w->attrib.override_redirect)
 	w->placed = TRUE;
+
+    if (initial)
+	damageWindowOutputExtents (w);
 }
 
 void
@@ -261,7 +261,7 @@ triggerButtonPressBindings (CompDisplay *d,
     unsigned int    bindMods;
     unsigned int    edge = 0;
 
-    if (event->xbutton.window == edgeWindow)
+    if (edgeWindow)
     {
 	CompScreen   *s;
 	unsigned int i;
@@ -270,6 +270,12 @@ triggerButtonPressBindings (CompDisplay *d,
 	if (!s)
 	    return FALSE;
 
+	if (event->xbutton.window != edgeWindow)
+	{
+	    if (!s->maxGrab || event->xbutton.window != s->root)
+		return FALSE;
+	}
+
 	for (i = 0; i < SCREEN_EDGE_NUM; i++)
 	{
 	    if (edgeWindow == s->screenEdge[i].id)
@@ -843,7 +849,7 @@ handleActionEvent (CompDisplay *d,
 
 		edgeWindow = None;
 
-		o[0].value.i = event->xcrossing.window;
+		o[0].value.i = d->activeWindow;
 		o[1].value.i = event->xcrossing.state;
 		o[2].value.i = event->xcrossing.x_root;
 		o[3].value.i = event->xcrossing.y_root;
@@ -887,7 +893,7 @@ handleActionEvent (CompDisplay *d,
 
 		edgeWindow = event->xcrossing.window;
 
-		o[0].value.i = event->xcrossing.window;
+		o[0].value.i = d->activeWindow;
 		o[1].value.i = event->xcrossing.state;
 		o[2].value.i = event->xcrossing.x_root;
 		o[3].value.i = event->xcrossing.y_root;
@@ -951,7 +957,7 @@ handleActionEvent (CompDisplay *d,
 	    {
 		state = CompActionStateTermEdgeDnd;
 
-		o[0].value.i = event->xclient.window;
+		o[0].value.i = d->activeWindow;
 		o[1].value.i = 0; /* fixme */
 		o[2].value.i = 0; /* fixme */
 		o[3].value.i = 0; /* fixme */
@@ -1006,7 +1012,7 @@ handleActionEvent (CompDisplay *d,
 	    {
 		state = CompActionStateInitEdgeDnd;
 
-		o[0].value.i = xdndWindow;
+		o[0].value.i = d->activeWindow;
 		o[1].value.i = 0; /* fixme */
 		o[2].value.i = event->xclient.data.l[2] >> 16;
 		o[3].value.i = event->xclient.data.l[2] & 0xffff;
@@ -1506,7 +1512,11 @@ handleEvent (CompDisplay *d,
 		finiTexture (s, &s->backgroundTexture);
 		initTexture (s, &s->backgroundTexture);
 
-		damageScreen (s);
+		if (s->backgroundLoaded)
+		{
+		    s->backgroundLoaded = FALSE;
+		    damageScreen (s);
+		}
 	    }
 	}
 	else if (event->xproperty.atom == d->wmStrutAtom ||
