@@ -63,13 +63,13 @@
 
 #define UNREDIRECT_FS_DEFAULT FALSE
 
-#define DEFAULT_ICON_DEFAULT "icon.png"
+#define DEFAULT_ICON_DEFAULT "icon"
 
 #define SYNC_TO_VBLANK_DEFAULT TRUE
 
 #define SCREEN_NUMBER_OF_DESKTOPS_DEFAULT 1
 #define SCREEN_NUMBER_OF_DESKTOPS_MIN     1
-#define SCREEN_NUMBER_OF_DESKTOPS_MAX     MAXSHORT
+#define SCREEN_NUMBER_OF_DESKTOPS_MAX     MAX_DESKTOPS
 
 #define DETECT_OUTPUTS_DEFAULT TRUE
 
@@ -217,44 +217,6 @@ setVirtualScreenSize (CompScreen *screen,
     setDesktopHints (screen);
 }
 
-static Bool
-updateDefaultIcon (CompScreen *screen)
-{
-    CompIcon     *icon;
-    char         *data;
-    unsigned int width, height;
-
-    if (!readPng (screen->opt[COMP_SCREEN_OPTION_DEFAULT_ICON].value.s,
-		  &data, &width, &height))
-	return FALSE;
-
-    icon = malloc (sizeof (CompIcon) + width * height * sizeof (CARD32));
-    if (!icon)
-    {
-	free (data);
-	return FALSE;
-    }
-
-    if (screen->defaultIcon)
-    {
-	finiTexture (screen, &screen->defaultIcon->texture);
-	free (screen->defaultIcon);
-    }
-
-    initTexture (screen, &icon->texture);
-
-    icon->width  = width;
-    icon->height = height;
-
-    memcpy (icon + 1, data, + width * height * sizeof (CARD32));
-
-    screen->defaultIcon = icon;
-
-    free (data);
-
-    return TRUE;
-}
-
 static void
 updateOutputDevices (CompScreen	*s)
 {
@@ -579,7 +541,8 @@ setScreenOption (CompScreen      *screen,
 	    return updateDefaultIcon (screen);
 	break;
     case COMP_SCREEN_OPTION_OUTPUTS:
-	if (screen->opt[COMP_SCREEN_OPTION_DETECT_OUTPUTS].value.b)
+	if (!noDetection &&
+	    screen->opt[COMP_SCREEN_OPTION_DETECT_OUTPUTS].value.b)
 	    return FALSE;
 
 	if (compSetOptionList (o, value))
@@ -849,14 +812,14 @@ updateScreenEdges (CompScreen *s)
 	int ww, w0;
 	int hh, h0;
     } geometry[SCREEN_EDGE_NUM] = {
-	{ 0,  0,   0,  1,   0,  1,   1, -2 }, /* left */
-	{ 1, -1,   0,  1,   0,  1,   1, -2 }, /* right */
-	{ 0,  1,   0,  0,   1, -2,   0,  1 }, /* top */
-	{ 0,  1,   1, -1,   1, -2,   0,  1 }, /* bottom */
-	{ 0,  0,   0,  0,   0,  1,   0,  1 }, /* top-left */
-	{ 1, -1,   0,  0,   0,  1,   0,  1 }, /* top-right */
-	{ 0,  0,   1, -1,   0,  1,   0,  1 }, /* bottom-left */
-	{ 1, -1,   1, -1,   0,  1,   0,  1 }  /* bottom-right */
+	{ 0,  0,   0,  2,   0,  2,   1, -4 }, /* left */
+	{ 1, -2,   0,  2,   0,  2,   1, -4 }, /* right */
+	{ 0,  2,   0,  0,   1, -4,   0,  2 }, /* top */
+	{ 0,  2,   1, -2,   1, -4,   0,  2 }, /* bottom */
+	{ 0,  0,   0,  0,   0,  2,   0,  2 }, /* top-left */
+	{ 1, -2,   0,  0,   0,  2,   0,  2 }, /* top-right */
+	{ 0,  0,   1, -2,   0,  2,   0,  2 }, /* bottom-left */
+	{ 1, -2,   1, -2,   0,  2,   0,  2 }  /* bottom-right */
     };
     int i;
 
@@ -930,6 +893,12 @@ reshape (CompScreen *s,
 	 int	    w,
 	 int	    h)
 {
+
+#ifdef USE_COW
+    if (useCow)
+	XMoveResizeWindow (s->display->display, s->overlay, 0, 0, w, h);
+#endif
+
     glMatrixMode (GL_PROJECTION);
     glLoadIdentity ();
     glMatrixMode (GL_MODELVIEW);
@@ -971,7 +940,7 @@ configureScreen (CompScreen	 *s,
 
 	reshape (s, ce->width, ce->height);
 
-	updateOutputDevices (s);
+	detectOutputDevices (s);
 
 	damageScreen (s);
     }
@@ -1077,7 +1046,7 @@ updateScreenBackground (CompScreen  *screen,
 	initTexture (screen, texture);
     }
 
-    if (!texture->name)
+    if (!texture->name && backgroundImage)
 	readImageToTexture (screen, texture, backgroundImage, &width, &height);
 
     if (texture->target == GL_TEXTURE_2D)
@@ -1326,27 +1295,76 @@ getDesktopHints (CompScreen *s)
 		     (unsigned char *) data, 1);
 }
 
+void
+showOutputWindow (CompScreen *s)
+{
+
 #ifdef USE_COW
+    if (useCow)
+    {
+	Display       *dpy = s->display->display;
+	XserverRegion region;
+
+	region = XFixesCreateRegion (dpy, NULL, 0);
+
+	XFixesSetWindowShapeRegion (dpy,
+				    s->output,
+				    ShapeBounding,
+				    0, 0, 0);
+	XFixesSetWindowShapeRegion (dpy,
+				    s->output,
+				    ShapeInput,
+				    0, 0, region);
+
+	XFixesDestroyRegion (dpy, region);
+
+	damageScreen (s);
+    }
+#endif
+
+}
+
+void
+hideOutputWindow (CompScreen *s)
+{
+
+#ifdef USE_COW
+    if (useCow)
+    {
+	Display       *dpy = s->display->display;
+	XserverRegion region;
+
+	region = XFixesCreateRegion (dpy, NULL, 0);
+
+	XFixesSetWindowShapeRegion (dpy,
+				    s->output,
+				    ShapeBounding,
+				    0, 0, region);
+
+	XFixesDestroyRegion (dpy, region);
+    }
+#endif
+
+}
+
 static void
 makeOutputWindow (CompScreen *s)
 {
-    Display       *dpy = s->display->display;
-    XserverRegion region;
 
-    s->overlay   = XCompositeGetOverlayWindow (dpy, s->root);
-    s->output    = s->overlay;
-
-    region = XFixesCreateRegion (dpy, NULL, 0);
-
-    XFixesSetWindowShapeRegion (dpy,
-				s->output,
-				ShapeInput,
-				0, 0, region);
-
-    XFixesDestroyRegion (dpy, region);
-}
+#ifdef USE_COW
+    if (useCow)
+    {
+	s->overlay = XCompositeGetOverlayWindow (s->display->display, s->root);
+	s->output  = s->overlay;
+    }
+    else
 #endif
 
+	s->output = s->overlay = s->root;
+
+    showOutputWindow (s);
+}
+
 Bool
 addScreen (CompDisplay *display,
 	   int	       screenNum,
@@ -1498,19 +1516,20 @@ addScreen (CompDisplay *display,
     s->initPluginForScreen = initPluginForScreen;
     s->finiPluginForScreen = finiPluginForScreen;
 
-    s->preparePaintScreen     = preparePaintScreen;
-    s->donePaintScreen        = donePaintScreen;
-    s->paintScreen	      = paintScreen;
-    s->paintTransformedScreen = paintTransformedScreen;
-    s->applyScreenTransform   = applyScreenTransform;
-    s->paintBackground        = paintBackground;
-    s->paintWindow            = paintWindow;
-    s->addWindowGeometry      = addWindowGeometry;
-    s->drawWindowTexture      = drawWindowTexture;
-    s->drawWindowGeometry     = drawWindowGeometry;
-    s->damageWindowRect       = damageWindowRect;
-    s->focusWindow	      = focusWindow;
-    s->setWindowScale	      = setWindowScale;
+    s->preparePaintScreen	 = preparePaintScreen;
+    s->donePaintScreen		 = donePaintScreen;
+    s->paintScreen		 = paintScreen;
+    s->paintTransformedScreen	 = paintTransformedScreen;
+    s->applyScreenTransform	 = applyScreenTransform;
+    s->paintBackground		 = paintBackground;
+    s->paintWindow		 = paintWindow;
+    s->drawWindow		 = drawWindow;
+    s->addWindowGeometry	 = addWindowGeometry;
+    s->drawWindowTexture	 = drawWindowTexture;
+    s->drawWindowGeometry	 = drawWindowGeometry;
+    s->damageWindowRect		 = damageWindowRect;
+    s->getOutputExtentsForWindow = getOutputExtentsForWindow;
+    s->focusWindow		 = focusWindow;
 
     s->windowResizeNotify = windowResizeNotify;
     s->windowMoveNotify	  = windowMoveNotify;
@@ -1533,13 +1552,7 @@ addScreen (CompDisplay *display,
 
     s->grabWindow = None;
 
-#ifdef USE_COW
-    if (useCow)
-	makeOutputWindow (s);
-    else
-#endif
-
-	s->overlay = s->output = s->root;
+    makeOutputWindow (s);
 
     templ.visualid = XVisualIDFromVisual (s->attrib.visual);
 
@@ -1600,23 +1613,50 @@ addScreen (CompDisplay *display,
     if (display->screens)
 	shareList = display->screens->ctx;
 
-    s->ctx = glXCreateContext (dpy, visinfo, shareList, !indirectRendering);
-    if (!s->ctx)
+    /* try both direct and indirect rendering contexts in case one of them
+       fail to support GLX_EXT_texture_from_pixmap */
+    for (i = 0; i < 2; i++)
     {
-	fprintf (stderr, "%s: glXCreateContext failed\n", programName);
-	return FALSE;
+	s->ctx = glXCreateContext (dpy, visinfo, shareList, !indirectRendering);
+	if (!s->ctx)
+	{
+	    fprintf (stderr, "%s: glXCreateContext failed\n", programName);
+	    XFree (visinfo);
+
+	    return FALSE;
+	}
+
+	if (glXIsDirect (dpy, s->ctx) == indirectRendering)
+	    i++;
+
+	glxExtensions = glXQueryExtensionsString (dpy, screenNum);
+	if (!strstr (glxExtensions, "GLX_EXT_texture_from_pixmap"))
+	{
+	    if (i > 0)
+	    {
+		fprintf (stderr, "%s: GLX_EXT_texture_from_pixmap is missing\n",
+			 programName);
+		XFree (visinfo);
+
+		return FALSE;
+	    }
+	    else
+	    {
+		fprintf (stderr, "%s: GLX_EXT_texture_from_pixmap is not "
+			 "supported by %s rendering context, trying %s "
+			 "rendering context instead\n", programName,
+			 indirectRendering ? "indirect" : "direct",
+			 indirectRendering ? "direct" : "indirect");
+
+		indirectRendering = !indirectRendering;
+
+		glXDestroyContext (dpy, s->ctx);
+	    }
+	}
     }
 
     XFree (visinfo);
 
-    glxExtensions = glXQueryExtensionsString (s->display->display, screenNum);
-    if (!strstr (glxExtensions, "GLX_EXT_texture_from_pixmap"))
-    {
-	fprintf (stderr, "%s: GLX_EXT_texture_from_pixmap is missing\n",
-		 programName);
-	return FALSE;
-    }
-
     if (!strstr (glxExtensions, "GLX_SGIX_fbconfig"))
     {
 	fprintf (stderr, "%s: GLX_SGIX_fbconfig is missing\n",
@@ -1929,14 +1969,9 @@ addScreen (CompDisplay *display,
     }
 
     initTexture (s, &s->backgroundTexture);
+    s->backgroundLoaded = FALSE;
 
     s->defaultIcon = NULL;
-    if (!updateDefaultIcon (s))
-    {
-	fprintf (stderr, "%s: Couldn't load default window icon.\n",
-		 programName);
-	return FALSE;
-    }
 
     s->desktopWindowCount = 0;
 
@@ -2018,9 +2053,11 @@ addScreen (CompDisplay *display,
     XFree (children);
 
     attrib.override_redirect = 1;
+    attrib.event_mask	     = PropertyChangeMask;
+
     s->grabWindow = XCreateWindow (dpy, s->root, -100, -100, 1, 1, 0,
-				   CopyFromParent, InputOnly,
-				   CopyFromParent, CWOverrideRedirect,
+				   CopyFromParent, InputOnly, CopyFromParent,
+				   CWOverrideRedirect | CWEventMask,
 				   &attrib);
     XMapWindow (dpy, s->grabWindow);
 
@@ -2702,13 +2739,26 @@ computeWorkareaForBox (CompScreen *s,
 		       XRectangle *area)
 {
     CompWindow *w;
+    Region     region;
+    REGION     r;
     int	       x1, y1, x2, y2;
-    int	       strutX1, strutY1, strutX2, strutY2;
 
-    strutX1 = pBox->x1;
-    strutY1 = pBox->y1;
-    strutX2 = pBox->x2;
-    strutY2 = pBox->y2;
+    region = XCreateRegion ();
+    if (!region)
+    {
+	area->x      = pBox->x1;
+	area->y      = pBox->y1;
+	area->width  = pBox->x1 - pBox->x1;
+	area->height = pBox->y2 - pBox->y1;
+
+	return;
+    }
+
+    r.rects    = &r.extents;
+    r.numRects = r.size = 1;
+    r.extents  = *pBox;
+
+    XUnionRegion (&r, region, region);
 
     for (w = s->windows; w; w = w->next)
     {
@@ -2717,6 +2767,9 @@ computeWorkareaForBox (CompScreen *s,
 
 	if (w->struts)
 	{
+	    r.extents.y1 = pBox->y1;
+	    r.extents.y2 = pBox->y2;
+
 	    x1 = w->struts->left.x;
 	    y1 = w->struts->left.y;
 	    x2 = x1 + w->struts->left.width;
@@ -2724,8 +2777,10 @@ computeWorkareaForBox (CompScreen *s,
 
 	    if (y1 < pBox->y2 && y2 > pBox->y1)
 	    {
-		if (x2 > strutX1)
-		    strutX1 = x2;
+		r.extents.x1 = x1;
+		r.extents.x2 = x2;
+
+		XSubtractRegion (region, &r, region);
 	    }
 
 	    x1 = w->struts->right.x;
@@ -2735,10 +2790,15 @@ computeWorkareaForBox (CompScreen *s,
 
 	    if (y1 < pBox->y2 && y2 > pBox->y1)
 	    {
-		if (x1 < strutX2)
-		    strutX2 = x1;
+		r.extents.x1 = x1;
+		r.extents.x2 = x2;
+
+		XSubtractRegion (region, &r, region);
 	    }
 
+	    r.extents.x1 = pBox->x1;
+	    r.extents.x2 = pBox->x2;
+
 	    x1 = w->struts->top.x;
 	    y1 = w->struts->top.y;
 	    x2 = x1 + w->struts->top.width;
@@ -2746,8 +2806,10 @@ computeWorkareaForBox (CompScreen *s,
 
 	    if (x1 < pBox->x2 && x2 > pBox->x1)
 	    {
-		if (y2 > strutY1)
-		    strutY1 = y2;
+		r.extents.y1 = y1;
+		r.extents.y2 = y2;
+
+		XSubtractRegion (region, &r, region);
 	    }
 
 	    x1 = w->struts->bottom.x;
@@ -2757,16 +2819,20 @@ computeWorkareaForBox (CompScreen *s,
 
 	    if (x1 < pBox->x2 && x2 > pBox->x1)
 	    {
-		if (y1 < strutY2)
-		    strutY2 = y1;
+		r.extents.y1 = y1;
+		r.extents.y2 = y2;
+
+		XSubtractRegion (region, &r, region);
 	    }
 	}
     }
 
-    area->x      = strutX1;
-    area->y      = strutY1;
-    area->width  = strutX2 - strutX1;
-    area->height = strutY2 - strutY1;
+    area->x      = region->extents.x1;
+    area->y      = region->extents.y1;
+    area->width  = region->extents.x2 - region->extents.x1;
+    area->height = region->extents.y2 - region->extents.y1;
+
+    XDestroyRegion (region);
 }
 
 void
@@ -3039,7 +3105,10 @@ moveScreenViewport (CompScreen *s,
 
     for (w = s->windows; w; w = w->next)
     {
-	if (!w->managed)
+	if (w->attrib.override_redirect)
+	    continue;
+
+	if (!w->managed && w->attrib.map_state != IsViewable)
 	    continue;
 
 	if (w->type & (CompWindowTypeDesktopMask | CompWindowTypeDockMask))
@@ -3422,9 +3491,9 @@ makeScreenCurrent (CompScreen *s)
     {
 	glXMakeCurrent (s->display->display, s->output, s->ctx);
 	currentRoot = s->root;
-
-	s->pendingCommands = TRUE;
     }
+
+    s->pendingCommands = TRUE;
 }
 
 void
@@ -3454,7 +3523,7 @@ outputDeviceForPoint (CompScreen *s,
 	x2 = s->outputDev[i].region.extents.x2;
 	y2 = s->outputDev[i].region.extents.y2;
 
-	if (x1 < x && x2 >= x && y1 < y && y2 >= y)
+	if (x1 <= x && x2 > x && y1 <= y && y2 > y)
 	    return i;
     }
 
@@ -3594,3 +3663,42 @@ clearScreenOutput (CompScreen	*s,
 	glClear (mask);
     }
 }
+
+Bool
+updateDefaultIcon (CompScreen *screen)
+{
+    CompIcon *icon;
+    char     *file = screen->opt[COMP_SCREEN_OPTION_DEFAULT_ICON].value.s;
+    void     *data;
+    int      width, height;
+
+    if (screen->defaultIcon)
+    {
+	finiTexture (screen, &screen->defaultIcon->texture);
+	free (screen->defaultIcon);
+	screen->defaultIcon = NULL;
+    }
+
+    if (!readImageFromFile (screen->display, file, &width, &height, &data))
+	return FALSE;
+
+    icon = malloc (sizeof (CompIcon) + width * height * sizeof (CARD32));
+    if (!icon)
+    {
+	free (data);
+	return FALSE;
+    }
+
+    initTexture (screen, &icon->texture);
+
+    icon->width  = width;
+    icon->height = height;
+
+    memcpy (icon + 1, data, + width * height * sizeof (CARD32));
+
+    screen->defaultIcon = icon;
+
+    free (data);
+
+    return TRUE;
+}
