@@ -556,6 +556,8 @@ setWindowActions (CompDisplay  *display,
 	data[i++] = display->winActionCloseAtom;
     if (actions & CompWindowActionShadeMask)
 	data[i++] = display->winActionShadeAtom;
+    if (actions & CompWindowActionChangeDesktopMask)
+	data[i++] = display->winActionChangeDesktopAtom;
 
     XChangeProperty (display->display, id, display->wmAllowedActionsAtom,
 		     XA_ATOM, 32, PropModeReplace,
@@ -578,7 +580,8 @@ recalcWindowActions (CompWindow *w)
 	    CompWindowActionResizeMask       |
 	    CompWindowActionStickMask        |
 	    CompWindowActionMinimizeMask     |
-	    CompWindowActionCloseMask;
+	    CompWindowActionCloseMask	     |
+	    CompWindowActionChangeDesktopMask;
 
 	if (w->input.top)
 	    actions |= CompWindowActionShadeMask;
@@ -589,7 +592,8 @@ recalcWindowActions (CompWindow *w)
 	    CompWindowActionMoveMask   |
 	    CompWindowActionResizeMask |
 	    CompWindowActionStickMask  |
-	    CompWindowActionCloseMask;
+	    CompWindowActionCloseMask  |
+	    CompWindowActionChangeDesktopMask;
 
 	if (w->input.top)
 	    actions |= CompWindowActionShadeMask;
@@ -600,7 +604,15 @@ recalcWindowActions (CompWindow *w)
 	    CompWindowActionMoveMask   |
 	    CompWindowActionResizeMask |
 	    CompWindowActionStickMask  |
-	    CompWindowActionCloseMask;
+	    CompWindowActionCloseMask  |
+	    CompWindowActionChangeDesktopMask;
+    default:
+	break;
+    }
+
+    switch (w->wmType) {
+    case CompWindowTypeNormalMask:
+	actions |= CompWindowActionFullscreenMask;
     default:
 	break;
     }
@@ -817,6 +829,48 @@ getProtocols (CompDisplay *display,
     return protocols;
 }
 
+unsigned int
+getWindowProp (CompDisplay  *display,
+	       Window	    id,
+	       Atom	    property,
+	       unsigned int defaultValue)
+{
+    Atom	  actual;
+    int		  result, format;
+    unsigned long n, left;
+    unsigned char *data;
+
+    result = XGetWindowProperty (display->display, id, property,
+				 0L, 1L, FALSE, XA_CARDINAL, &actual, &format,
+				 &n, &left, &data);
+
+    if (result == Success && n && data)
+    {
+	unsigned long value;
+
+	memcpy (&value, data, sizeof (unsigned long));
+
+	XFree (data);
+
+	return (unsigned int) value;
+    }
+
+    return defaultValue;
+}
+
+void
+setWindowProp (CompDisplay  *display,
+	       Window       id,
+	       Atom	    property,
+	       unsigned int value)
+{
+    unsigned long data = value;
+
+    XChangeProperty (display->display, id, property,
+		     XA_CARDINAL, 32, PropModeReplace,
+		     (unsigned char *) &data, 1);
+}
+
 unsigned short
 getWindowProp32 (CompDisplay	*display,
 		 Window		id,
@@ -861,7 +915,6 @@ setWindowProp32 (CompDisplay    *display,
 		     (unsigned char *) &value32, 1);
 }
 
-
 static void
 updateFrameWindow (CompWindow *w)
 {
@@ -870,11 +923,12 @@ updateFrameWindow (CompWindow *w)
 	XRectangle rects[4];
 	int	   x, y, width, height;
 	int	   i = 0;
+	int	   bw = w->serverBorderWidth * 2;
 
 	x      = w->serverX - w->input.left;
 	y      = w->serverY - w->input.top;
-	width  = w->width  + w->input.left + w->input.right;
-	height = w->height + w->input.top  + w->input.bottom;
+	width  = w->serverWidth  + w->input.left + w->input.right + bw;
+	height = w->serverHeight + w->input.top  + w->input.bottom + bw;
 
 	if (!w->frame)
 	{
@@ -970,8 +1024,7 @@ updateFrameWindow (CompWindow *w)
 
 void
 setWindowFrameExtents (CompWindow	 *w,
-		       CompWindowExtents *input,
-		       CompWindowExtents *output)
+		       CompWindowExtents *input)
 {
     if (input->left   != w->input.left  ||
 	input->right  != w->input.right ||
@@ -987,6 +1040,7 @@ setWindowFrameExtents (CompWindow	 *w,
 	data[2] = input->top;
 	data[3] = input->bottom;
 
+	updateWindowSize (w);
 	updateFrameWindow (w);
 
 	XChangeProperty (w->screen->display->display, w->id,
@@ -994,13 +1048,21 @@ setWindowFrameExtents (CompWindow	 *w,
 			 XA_CARDINAL, 32, PropModeReplace,
 			 (unsigned char *) data, 4);
     }
+}
 
-    if (output->left   != w->output.left  ||
-	output->right  != w->output.right ||
-	output->top    != w->output.top   ||
-	output->bottom != w->output.bottom)
+void
+updateWindowOutputExtents (CompWindow *w)
+{
+    CompWindowExtents output;
+
+    (*w->screen->getOutputExtentsForWindow) (w, &output);
+
+    if (output.left   != w->output.left  ||
+	output.right  != w->output.right ||
+	output.top    != w->output.top   ||
+	output.bottom != w->output.bottom)
     {
-	w->output = *output;
+	w->output = output;
 
 	(*w->screen->windowResizeNotify) (w);
     }
@@ -1132,33 +1194,36 @@ freeWindow (CompWindow *w)
 }
 
 void
-damageWindowRegion (CompWindow *w,
-		    Region     region)
+damageTransformedWindowRect (CompWindow *w,
+			     float	xScale,
+			     float	yScale,
+			     float	xTranslate,
+			     float	yTranslate,
+			     BoxPtr     rect)
 {
-    if (w->scaled)
+    REGION reg;
+
+    reg.rects    = &reg.extents;
+    reg.numRects = 1;
+
+    reg.extents.x1 = (rect->x1 * xScale);
+    reg.extents.y1 = (rect->y1 * yScale);
+    reg.extents.x2 = (rect->x2 * xScale + 0.5f);
+    reg.extents.y2 = (rect->y2 * yScale + 0.5f);
+
+    reg.extents.x1 += xTranslate;
+    reg.extents.y1 += yTranslate;
+    reg.extents.x2 += (xTranslate + 0.5f);
+    reg.extents.y2 += (yTranslate + 0.5f);
+
+    if (reg.extents.x2 > reg.extents.x1 && reg.extents.y2 > reg.extents.y1)
     {
-	REGION reg;
-	int    x1, y1, x2, y2;
+	reg.extents.x1 += w->attrib.x + w->attrib.border_width;
+	reg.extents.y1 += w->attrib.y + w->attrib.border_width;
+	reg.extents.x2 += w->attrib.x + w->attrib.border_width;
+	reg.extents.y2 += w->attrib.y + w->attrib.border_width;
 
-	reg.rects    = &reg.extents;
-	reg.numRects = 1;
-
-	x1 = region->extents.x1 - w->attrib.x;
-	y1 = region->extents.y1 - w->attrib.y;
-	x2 = region->extents.x2 - w->attrib.x;
-	y2 = region->extents.y2 - w->attrib.y;
-
-	reg.extents.x1 = (x1 * w->paint.xScale) + w->attrib.x;
-	reg.extents.y1 = (y1 * w->paint.yScale) + w->attrib.y;
-	reg.extents.x2 = (x2 * w->paint.xScale + 0.5f) + w->attrib.x;
-	reg.extents.y2 = (y2 * w->paint.yScale + 0.5f) + w->attrib.y;
-
-	if (reg.extents.x2 > reg.extents.x1 && reg.extents.y2 > reg.extents.y1)
-	    damageScreenRegion (w->screen, &reg);
-    }
-    else
-    {
-	damageScreenRegion (w->screen, region);
+	damageScreenRegion (w->screen, &reg);
     }
 }
 
@@ -1170,42 +1235,39 @@ damageWindowOutputExtents (CompWindow *w)
 
     if (w->shaded || (w->attrib.map_state == IsViewable && w->damaged))
     {
-	REGION reg;
-
-	reg.rects = &reg.extents;
-	reg.numRects = reg.size = 1;
+	BoxRec box;
 
 	/* top */
-	reg.extents.x1 = w->attrib.x - w->output.left;
-	reg.extents.y1 = w->attrib.y - w->output.top;
-	reg.extents.x2 = w->attrib.x + w->width + w->output.right;
-	reg.extents.y2 = w->attrib.y;
+	box.x1 = -w->output.left;
+	box.y1 = -w->output.top;
+	box.x2 = w->width + w->output.right;
+	box.y2 = 0;
 
-	if (reg.extents.x1 < reg.extents.x2 && reg.extents.y1 < reg.extents.y2)
-	    damageWindowRegion (w, &reg);
+	if (box.x1 < box.x2 && box.y1 < box.y2)
+	    addWindowDamageRect (w, &box);
 
 	/* bottom */
-	reg.extents.y1 = w->attrib.y + w->height;
-	reg.extents.y2 = reg.extents.y1 + w->output.bottom;
+	box.y1 = w->height;
+	box.y2 = box.y1 + w->output.bottom;
 
-	if (reg.extents.x1 < reg.extents.x2 && reg.extents.y1 < reg.extents.y2)
-	    damageWindowRegion (w, &reg);
+	if (box.x1 < box.x2 && box.y1 < box.y2)
+	    addWindowDamageRect (w, &box);
 
 	/* left */
-	reg.extents.x1 = w->attrib.x - w->output.left;
-	reg.extents.y1 = w->attrib.y;
-	reg.extents.x2 = w->attrib.x;
-	reg.extents.y2 = w->attrib.y + w->height;
+	box.x1 = -w->output.left;
+	box.y1 = 0;
+	box.x2 = 0;
+	box.y2 = w->height;
 
-	if (reg.extents.x1 < reg.extents.x2 && reg.extents.y1 < reg.extents.y2)
-	    damageWindowRegion (w, &reg);
+	if (box.x1 < box.x2 && box.y1 < box.y2)
+	    addWindowDamageRect (w, &box);
 
 	/* right */
-	reg.extents.x1 = w->attrib.x + w->width;
-	reg.extents.x2 = reg.extents.x1 + w->output.right;
+	box.x1 = w->width;
+	box.x2 = box.x1 + w->output.right;
 
-	if (reg.extents.x1 < reg.extents.x2 && reg.extents.y1 < reg.extents.y2)
-	    damageWindowRegion (w, &reg);
+	if (box.x1 < box.x2 && box.y1 < box.y2)
+	    addWindowDamageRect (w, &box);
     }
 }
 
@@ -1217,6 +1279,41 @@ damageWindowRect (CompWindow *w,
     return FALSE;
 }
 
+void
+addWindowDamageRect (CompWindow *w,
+		     BoxPtr     rect)
+{
+    REGION region;
+
+    if (w->screen->damageMask & COMP_SCREEN_DAMAGE_ALL_MASK)
+	return;
+
+    region.extents = *rect;
+
+    if (!(*w->screen->damageWindowRect) (w, FALSE, &region.extents))
+    {
+	region.extents.x1 += w->attrib.x + w->attrib.border_width;
+	region.extents.y1 += w->attrib.y + w->attrib.border_width;
+	region.extents.x2 += w->attrib.x + w->attrib.border_width;
+	region.extents.y2 += w->attrib.y + w->attrib.border_width;
+
+	region.rects = &region.extents;
+	region.numRects = region.size = 1;
+
+	damageScreenRegion (w->screen, &region);
+    }
+}
+
+void
+getOutputExtentsForWindow (CompWindow	     *w,
+			   CompWindowExtents *output)
+{
+    output->left   = 0;
+    output->right  = 0;
+    output->top    = 0;
+    output->bottom = 0;
+}
+
 void
 addWindowDamage (CompWindow *w)
 {
@@ -1225,25 +1322,14 @@ addWindowDamage (CompWindow *w)
 
     if (w->shaded || (w->attrib.map_state == IsViewable && w->damaged))
     {
-	REGION region;
+	BoxRec box;
 
-	region.extents.x1 = -w->output.left - w->attrib.border_width;
-	region.extents.y1 = -w->output.top - w->attrib.border_width;
-	region.extents.x2 = w->width + w->output.right;
-	region.extents.y2 = w->height + w->output.bottom;
+	box.x1 = -w->output.left - w->attrib.border_width;
+	box.y1 = -w->output.top - w->attrib.border_width;
+	box.x2 = w->width + w->output.right;
+	box.y2 = w->height + w->output.bottom;
 
-	if (!(*w->screen->damageWindowRect) (w, FALSE, &region.extents))
-	{
-	    region.extents.x1 += w->attrib.x + w->attrib.border_width;
-	    region.extents.y1 += w->attrib.y + w->attrib.border_width;
-	    region.extents.x2 += w->attrib.x + w->attrib.border_width;
-	    region.extents.y2 += w->attrib.y + w->attrib.border_width;
-
-	    region.rects = &region.extents;
-	    region.numRects = region.size = 1;
-
-	    damageWindowRegion (w, &region);
-	}
+	addWindowDamageRect (w, &box);
     }
 }
 
@@ -1443,6 +1529,58 @@ updateWindowStruts (CompWindow *w)
 	}
     }
 
+    if (hasNew)
+    {
+	int strutX1, strutY1, strutX2, strutY2;
+	int x1, y1, x2, y2;
+	int i;
+
+	/* applications expect us to clip struts to xinerama edges */
+	for (i = 0; i < w->screen->display->nScreenInfo; i++)
+	{
+	    x1 = w->screen->display->screenInfo[i].x_org;
+	    y1 = w->screen->display->screenInfo[i].y_org;
+	    x2 = x1 + w->screen->display->screenInfo[i].width;
+	    y2 = y1 + w->screen->display->screenInfo[i].height;
+
+	    strutX1 = new.left.x;
+	    strutX2 = strutX1 + new.left.width;
+
+	    if (strutX2 > x1 && strutX2 <= x2)
+	    {
+		new.left.x     = x1;
+		new.left.width = strutX2 - x1;
+	    }
+
+	    strutX1 = new.right.x;
+	    strutX2 = strutX1 + new.right.width;
+
+	    if (strutX1 > x1 && strutX1 <= x2)
+	    {
+		new.right.x     = strutX1;
+		new.right.width = x2 - strutX1;
+	    }
+
+	    strutY1 = new.top.y;
+	    strutY2 = strutY1 + new.top.height;
+
+	    if (strutY2 > y1 && strutY2 <= y2)
+	    {
+		new.top.y      = y1;
+		new.top.height = strutY2 - y1;
+	    }
+
+	    strutY1 = new.bottom.y;
+	    strutY2 = strutY1 + new.bottom.height;
+
+	    if (strutY1 > y1 && strutY1 <= y2)
+	    {
+		new.bottom.y      = strutY1;
+		new.bottom.height = y2 - strutY1;
+	    }
+	}
+    }
+
     if (hasOld != hasNew || (hasNew && hasOld &&
 			     memcmp (&new, &old, sizeof (CompStruts))))
     {
@@ -1518,10 +1656,13 @@ addWindow (CompScreen *screen,
     w->shaded		 = FALSE;
     w->hidden		 = FALSE;
 
+    w->desktop = 0xffffffff;
+
     w->initialViewportX = screen->x;
     w->initialViewportY = screen->y;
 
     w->pendingUnmaps = 0;
+    w->pendingMaps   = 0;
 
     w->startupId = NULL;
     w->resName   = NULL;
@@ -1578,13 +1719,13 @@ addWindow (CompScreen *screen,
     w->paint.saturation = w->saturation = COLOR;
     w->paint.xScale	= 1.0f;
     w->paint.yScale	= 1.0f;
+    w->paint.xTranslate	= 0.0f;
+    w->paint.yTranslate	= 0.0f;
 
     w->lastPaint = w->paint;
 
     w->alive = TRUE;
 
-    w->scaled = FALSE;
-
     w->mwmDecor = MwmDecorAll;
     w->mwmFunc  = MwmFuncAll;
 
@@ -1720,13 +1861,7 @@ addWindow (CompScreen *screen,
     {
 	updateNormalHints (w);
 	updateWindowStruts (w);
-
 	updateWmHints (w);
-
-	setWindowProp32 (w->screen->display, w->id,
-			 w->screen->display->winDesktopAtom,
-			 0);
-
 	updateTransientHint (w);
 
 	w->clientLeader = getClientLeader (w);
@@ -1739,16 +1874,27 @@ addWindow (CompScreen *screen,
 
 	recalcWindowActions (w);
 
-	if (!(w->type & CompWindowTypeDesktopMask))
-	    w->opacity =
-		getWindowProp32 (w->screen->display, w->id,
-				 w->screen->display->winOpacityAtom,
-				 OPAQUE);
+	if (!(w->type & (CompWindowTypeDesktopMask | CompWindowTypeDockMask)))
+	{
+	    unsigned int desktop;
+
+	    desktop = getWindowProp (w->screen->display, w->id,
+				     w->screen->display->winDesktopAtom,
+				     w->screen->currentDesktop);
+	    if (desktop >= 0 && desktop < w->screen->nDesktop)
+		w->desktop = desktop;
+
+	    if (!(w->type & CompWindowTypeDesktopMask))
+		w->opacity =
+		    getWindowProp32 (w->screen->display, w->id,
+				     w->screen->display->winOpacityAtom,
+				     OPAQUE);
+	}
 
 	w->brightness =
-		getWindowProp32 (w->screen->display, w->id,
-				 w->screen->display->winBrightnessAtom,
-				 BRIGHT);
+	    getWindowProp32 (w->screen->display, w->id,
+			     w->screen->display->winBrightnessAtom,
+			     BRIGHT);
 
 	if (w->alive)
 	{
@@ -1776,8 +1922,23 @@ addWindow (CompScreen *screen,
 	w->attrib.map_state = IsUnmapped;
 
 	if (!w->attrib.override_redirect)
+	{
 	    w->managed = TRUE;
 
+	    if (w->wmType & (CompWindowTypeDockMask |
+			     CompWindowTypeDesktopMask))
+		setDesktopForWindow (w, 0xffffffff);
+
+	    if (w->desktop != 0xffffffff)
+		w->desktop = w->screen->currentDesktop;
+
+	    setWindowProp (w->screen->display, w->id,
+			   w->screen->display->winDesktopAtom,
+			   w->desktop);
+	}
+
+	w->pendingMaps++;
+
 	mapWindow (w);
 
 	updateWindowAttributes (w, FALSE);
@@ -1788,10 +1949,13 @@ addWindow (CompScreen *screen,
 	{
 	    w->managed = TRUE;
 
-	    if (w->state & CompWindowStateShadedMask)
-		w->shaded = TRUE;
-	    else
-		w->minimized = TRUE;
+	    if (w->state & CompWindowStateHiddenMask)
+	    {
+		if (w->state & CompWindowStateShadedMask)
+		    w->shaded = TRUE;
+		else
+		    w->minimized = TRUE;
+	    }
 	}
     }
 
@@ -1816,17 +1980,18 @@ removeWindow (CompWindow *w)
 
 	if (w->struts)
 	    updateWorkareaForScreen (w->screen);
+    }
 
-	updateClientListForScreen (w->screen);
-    }
-    else if (w->state & CompWindowStateHiddenMask)
-    {
-	updateClientListForScreen (w->screen);
-    }
+    updateClientListForScreen (w->screen);
 
     if (!w->redirected)
+    {
 	w->screen->overlayWindowCount--;
 
+	if (w->screen->overlayWindowCount < 1)
+	    showOutputWindow (w->screen);
+    }
+
     windowFiniPlugins (w);
 
     freeWindow (w);
@@ -1854,9 +2019,9 @@ sendConfigureNotify (CompWindow *w)
 {
     XConfigureEvent xev;
 
-    xev.type	     = ConfigureNotify;
-    xev.event	     = w->id;
-    xev.window	     = w->id;
+    xev.type   = ConfigureNotify;
+    xev.event  = w->id;
+    xev.window = w->id;
 
     /* normally we should never send configure notify events to override
        redirect windows but if they support the _NET_WM_SYNC_REQUEST
@@ -1877,7 +2042,7 @@ sendConfigureNotify (CompWindow *w)
 	    xev.height	     = attrib.height;
 	    xev.border_width = attrib.border_width;
 
-	    xev.above	      = (w->prev) ? w->prev->id : None;
+	    xev.above		  = (w->prev) ? w->prev->id : None;
 	    xev.override_redirect = TRUE;
 
 	    XSendEvent (w->screen->display->display, w->id, FALSE,
@@ -1888,13 +2053,13 @@ sendConfigureNotify (CompWindow *w)
     }
     else
     {
-	xev.x	     = w->attrib.x;
-	xev.y	     = w->attrib.y;
-	xev.width	     = w->serverWidth;
-	xev.height	     = w->serverHeight;
+	xev.x		 = w->attrib.x;
+	xev.y		 = w->attrib.y;
+	xev.width	 = w->serverWidth;
+	xev.height	 = w->serverHeight;
 	xev.border_width = w->serverBorderWidth;
 
-	xev.above	          = (w->prev) ? w->prev->id : None;
+	xev.above	      = (w->prev) ? w->prev->id : None;
 	xev.override_redirect = w->attrib.override_redirect;
 
 	XSendEvent (w->screen->display->display, w->id, FALSE,
@@ -1908,8 +2073,13 @@ mapWindow (CompWindow *w)
     if (w->attrib.map_state == IsViewable)
 	return;
 
+    w->pendingMaps--;
+
     w->mapNum = w->screen->mapNum++;
 
+    if (w->struts)
+	updateWorkareaForScreen (w->screen);
+
     if (w->attrib.class == InputOnly)
 	return;
 
@@ -1932,9 +2102,6 @@ mapWindow (CompWindow *w)
     if (w->frame)
 	XMapWindow (w->screen->display->display, w->frame);
 
-    if (w->struts)
-	updateWorkareaForScreen (w->screen);
-
     updateClientListForScreen (w->screen);
 
     if (w->type & CompWindowTypeDesktopMask)
@@ -1942,8 +2109,14 @@ mapWindow (CompWindow *w)
 
     if (w->protocols & CompWindowProtocolSyncRequestMask)
     {
-	sendSyncRequest (w);
-	sendConfigureNotify (w);
+	/* XXX: override_redirect windows and _NET_WM_SYNC_REQUEST
+	   protocol is causing issues. Avoid sending sync requests to
+	   override_redirect windows until this is solved. */
+	if (!w->attrib.override_redirect)
+	{
+	    sendSyncRequest (w);
+	    sendConfigureNotify (w);
+	}
     }
 
     if (!w->attrib.override_redirect)
@@ -1972,6 +2145,9 @@ unmapWindow (CompWindow *w)
     if (w->unmapRefCnt > 0)
 	return;
 
+    if (w->struts)
+	updateWorkareaForScreen (w->screen);
+
     if (w->attrib.map_state != IsViewable)
 	return;
 
@@ -1992,9 +2168,6 @@ unmapWindow (CompWindow *w)
 		      w->attrib.width, ++w->attrib.height - 1,
 		      w->attrib.border_width);
 
-    if (w->struts)
-	updateWorkareaForScreen (w->screen);
-
     updateClientListForScreen (w->screen);
 }
 
@@ -2333,35 +2506,19 @@ syncWindowPosition (CompWindow *w)
 		     w->attrib.y - w->input.top);
 }
 
-void
-setWindowScale (CompWindow *w,
-		float      xScale,
-		float      yScale)
-{
-    if (xScale > 0.999f && xScale < 1.001f &&
-	yScale > 0.999f && yScale < 1.001f)
-    {
-	w->paint.xScale = 1.0f;
-	w->paint.yScale = 1.0f;
-
-	w->scaled = FALSE;
-    }
-    else
-    {
-	w->paint.xScale = xScale;
-	w->paint.yScale = yScale;
-
-	w->scaled = TRUE;
-    }
-}
-
 Bool
 focusWindow (CompWindow *w)
 {
     if (w->attrib.override_redirect)
 	return FALSE;
 
-    if (!w->shaded && (!w->mapNum || w->attrib.map_state != IsViewable))
+    if (!w->managed)
+	return FALSE;
+
+    if (!onCurrentDesktop (w))
+	return FALSE;
+
+    if (!w->shaded && (w->state & CompWindowStateHiddenMask))
 	return FALSE;
 
     if (w->attrib.x + w->width  <= 0	||
@@ -2525,6 +2682,20 @@ moveInputFocusToWindow (CompWindow *w)
 
 	    XSendEvent (d->display, w->id, FALSE, NoEventMask, &ev);
 	}
+	else if (!modalTransient)
+	{
+	    CompWindow *ancestor;
+
+	    /* move input to closest ancestor */
+	    for (ancestor = s->windows; ancestor; ancestor = ancestor->next)
+	    {
+		if (isAncestorTo (w, ancestor))
+		{
+		    moveInputFocusToWindow (ancestor);
+		    break;
+		}
+	    }
+	}
     }
 }
 
@@ -2566,7 +2737,7 @@ avoidStackingRelativeTo (CompWindow *w)
     if (w->attrib.override_redirect)
 	return TRUE;
 
-    if (!w->shaded)
+    if (!w->shaded && !w->pendingMaps)
     {
 	if (w->attrib.map_state != IsViewable || w->mapNum == 0)
 	    return TRUE;
@@ -2891,7 +3062,7 @@ stackTransients (CompWindow	*w,
 	    if (xwc->sibling == t->id)
 		return FALSE;
 
-	    if (t->mapNum)
+	    if (t->mapNum || t->pendingMaps)
 		configureXWindow (t, CWSibling | CWStackMode, xwc);
 	}
     }
@@ -2916,7 +3087,7 @@ stackAncestors (CompWindow     *w,
 	    if (ancestor->type & CompWindowTypeDesktopMask)
 		return;
 
-	    if (ancestor->mapNum)
+	    if (ancestor->mapNum || ancestor->pendingMaps)
 		configureXWindow (ancestor,
 				  CWSibling | CWStackMode,
 				  xwc);
@@ -2943,7 +3114,7 @@ stackAncestors (CompWindow     *w,
 		if (a->type & CompWindowTypeDesktopMask)
 		    continue;
 
-		if (a->mapNum)
+		if (a->mapNum || a->pendingMaps)
 		    configureXWindow (a,
 				      CWSibling | CWStackMode,
 				      xwc);
@@ -2961,21 +3132,26 @@ addWindowSizeChanges (CompWindow     *w,
 		      int	     oldHeight,
 		      int	     oldBorderWidth)
 {
-    int mask = 0;
-    int x, y;
-    int vx, vy;
+    XRectangle workArea;
+    int	       mask = 0;
+    int	       x, y;
+    int	       vx, vy;
+    int	       output;
 
     defaultViewportForWindow (w, &vx, &vy);
 
     x = (vx - w->screen->x) * w->screen->width;
     y = (vy - w->screen->y) * w->screen->height;
 
+    output = outputDeviceForWindow (w);
+    getWorkareaForOutput (w->screen, output, &workArea);
+
     if (w->type & CompWindowTypeFullscreenMask)
     {
 	saveWindowGeometry (w, CWX | CWY | CWWidth | CWHeight | CWBorderWidth);
 
-	xwc->width	  = w->screen->width;
-	xwc->height	  = w->screen->height;
+	xwc->width	  = w->screen->outputDev[output].width;
+	xwc->height	  = w->screen->outputDev[output].height;
 	xwc->border_width = 0;
 
 	mask |= CWWidth | CWHeight | CWBorderWidth;
@@ -2988,7 +3164,7 @@ addWindowSizeChanges (CompWindow     *w,
 	{
 	    saveWindowGeometry (w, CWY | CWHeight);
 
-	    xwc->height = w->screen->workArea.height - w->input.top -
+	    xwc->height = workArea.height - w->input.top -
 		w->input.bottom - oldBorderWidth * 2;
 
 	    mask |= CWHeight;
@@ -3002,7 +3178,7 @@ addWindowSizeChanges (CompWindow     *w,
 	{
 	    saveWindowGeometry (w, CWX | CWWidth);
 
-	    xwc->width = w->screen->workArea.width - w->input.left -
+	    xwc->width = workArea.width - w->input.left -
 		w->input.right - oldBorderWidth * 2;
 
 	    mask |= CWWidth;
@@ -3031,8 +3207,8 @@ addWindowSizeChanges (CompWindow     *w,
     {
 	if (w->type & CompWindowTypeFullscreenMask)
 	{
-	    xwc->x = x;
-	    xwc->y = y;
+	    xwc->x = x + w->screen->outputDev[output].region.extents.x1;
+	    xwc->y = y + w->screen->outputDev[output].region.extents.y1;
 
 	    mask |= CWX | CWY;
 	}
@@ -3056,16 +3232,16 @@ addWindowSizeChanges (CompWindow     *w,
 
 	    if (w->state & CompWindowStateMaximizedVertMask)
 	    {
-		if (oldY < y + w->screen->workArea.y + w->input.top)
+		if (oldY < y + workArea.y + w->input.top)
 		{
-		    xwc->y = y + w->screen->workArea.y + w->input.top;
+		    xwc->y = y + workArea.y + w->input.top;
 		    mask |= CWY;
 		}
 		else
 		{
 		    height = xwc->height + oldBorderWidth * 2;
 
-		    max = y + w->screen->workArea.y + w->screen->workArea.height;
+		    max = y + workArea.y + workArea.height;
 		    if (oldY + oldHeight + w->input.bottom > max)
 		    {
 			xwc->y = max - height - w->input.bottom;
@@ -3073,9 +3249,9 @@ addWindowSizeChanges (CompWindow     *w,
 		    }
 		    else if (oldY + height + w->input.bottom > max)
 		    {
-			xwc->y = w->screen->workArea.y +
-			    (w->screen->workArea.height - w->input.top -
-			     height - w->input.bottom) / 2 + w->input.top;
+			xwc->y = y + workArea.y +
+			    (workArea.height - w->input.top - height -
+			     w->input.bottom) / 2 + w->input.top;
 			mask |= CWY;
 		    }
 		}
@@ -3083,16 +3259,16 @@ addWindowSizeChanges (CompWindow     *w,
 
 	    if (w->state & CompWindowStateMaximizedHorzMask)
 	    {
-		if (oldX < x + w->screen->workArea.x + w->input.left)
+		if (oldX < x + workArea.x + w->input.left)
 		{
-		    xwc->x = x + w->screen->workArea.x + w->input.left;
+		    xwc->x = x + workArea.x + w->input.left;
 		    mask |= CWX;
 		}
 		else
 		{
 		    width = xwc->width + oldBorderWidth * 2;
 
-		    max = x + w->screen->workArea.x + w->screen->workArea.width;
+		    max = x + workArea.x + workArea.width;
 		    if (oldX + oldWidth + w->input.right > max)
 		    {
 			xwc->x = max - width - w->input.right;
@@ -3100,9 +3276,9 @@ addWindowSizeChanges (CompWindow     *w,
 		    }
 		    else if (oldX + width + w->input.right > max)
 		    {
-			xwc->x = w->screen->workArea.x +
-			    (w->screen->workArea.width - w->input.left -
-			     width - w->input.right) / 2 + w->input.left;
+			xwc->x = x + workArea.x +
+			    (workArea.width - w->input.left - width -
+			     w->input.right) / 2 + w->input.left;
 			mask |= CWX;
 		    }
 		}
@@ -3217,6 +3393,9 @@ moveResizeWindow (CompWindow     *w,
 	    min = w->screen->workArea.y + w->input.top;
 	    max = w->screen->workArea.y + w->screen->workArea.height;
 
+	    min -= w->screen->y * w->screen->height;
+	    max += (w->screen->vsize - w->screen->y - 1) * w->screen->height;
+
 	    if (xwc->y < min)
 		xwc->y = min;
 	    else if (xwc->y > max)
@@ -3230,6 +3409,9 @@ moveResizeWindow (CompWindow     *w,
 	    min = w->screen->workArea.x + w->input.left;
 	    max = w->screen->workArea.x + w->screen->workArea.width;
 
+	    min -= w->screen->x * w->screen->width;
+	    max += (w->screen->hsize - w->screen->x - 1) * w->screen->width;
+
 	    if (xwc->x < min)
 		xwc->x = min;
 	    else if (xwc->x > max)
@@ -3273,6 +3455,9 @@ updateWindowSize (CompWindow *w)
     XWindowChanges xwc;
     int		   mask;
 
+    if (w->attrib.override_redirect || !w->managed)
+	return;
+
     mask = addWindowSizeChanges (w, &xwc,
 				 w->attrib.x, w->attrib.y,
 				 w->serverWidth, w->serverHeight,
@@ -3491,16 +3676,38 @@ ensureWindowVisibility (CompWindow *w)
     }
 }
 
+static void
+revealWindow (CompWindow *w)
+{
+    if (w->minimized)
+	unminimizeWindow (w);
+
+    leaveShowDesktopMode (w->screen, w);
+}
+
+static void
+revealAncestors (CompWindow *w,
+		 void       *closure)
+{
+    CompWindow *transient = closure;
+
+    if (isAncestorTo (transient, w))
+    {
+	forEachWindowOnScreen (w->screen, revealAncestors, (void *) w);
+	revealWindow (w);
+    }
+}
+
 void
 activateWindow (CompWindow *w)
 {
+    setCurrentDesktop (w->screen, w->desktop);
+
+    forEachWindowOnScreen (w->screen, revealAncestors, (void *) w);
+    revealWindow (w);
+
     if (w->state & CompWindowStateHiddenMask)
     {
-	if (w->minimized)
-	    unminimizeWindow (w);
-
-	leaveShowDesktopMode (w->screen, w);
-
 	w->state &= ~CompWindowStateShadedMask;
 	if (w->shaded)
 	    showWindow (w);
@@ -3509,6 +3716,9 @@ activateWindow (CompWindow *w)
     if (w->state & CompWindowStateHiddenMask)
 	return;
 
+    if (!onCurrentDesktop (w))
+	return;
+
     ensureWindowVisibility (w);
     updateWindowAttributes (w, TRUE);
     moveInputFocusToWindow (w);
@@ -3520,6 +3730,9 @@ closeWindow (CompWindow *w,
 {
     CompDisplay *display = w->screen->display;
 
+    if (serverTime == 0)
+	serverTime = getCurrentTimeFromDisplay (display);
+
     if (w->alive)
     {
 	if (w->protocols & CompWindowProtocolDeleteMask)
@@ -3714,7 +3927,12 @@ constrainNewWindowSize (CompWindow *w,
 void
 hideWindow (CompWindow *w)
 {
-    if (!w->minimized && !w->inShowDesktopMode && !w->hidden)
+    Bool onDesktop = onCurrentDesktop (w);
+
+    if (!w->managed)
+	return;
+
+    if (!w->minimized && !w->inShowDesktopMode && !w->hidden && onDesktop)
     {
 	if (w->state & CompWindowStateShadedMask)
 	{
@@ -3738,10 +3956,8 @@ hideWindow (CompWindow *w)
     if (w->attrib.map_state != IsViewable)
 	return;
 
-    if (w->state & CompWindowStateHiddenMask)
-	return;
-
-    w->state |= CompWindowStateHiddenMask;
+    if (w->minimized || w->inShowDesktopMode || w->hidden || w->shaded)
+	w->state |= CompWindowStateHiddenMask;
 
     w->pendingUnmaps++;
 
@@ -3758,11 +3974,28 @@ hideWindow (CompWindow *w)
 void
 showWindow (CompWindow *w)
 {
-    if (!(w->state & CompWindowStateHiddenMask))
+    Bool onDesktop = onCurrentDesktop (w);
+
+    if (!w->managed)
 	return;
 
-    if (w->minimized || w->inShowDesktopMode || w->hidden)
+    if (w->minimized || w->inShowDesktopMode || w->hidden || !onDesktop)
+    {
+	/* no longer hidden but not on current desktop */
+	if (!w->minimized && !w->inShowDesktopMode && !w->hidden)
+	{
+	    if (w->state & CompWindowStateHiddenMask)
+	    {
+		w->state &= ~CompWindowStateHiddenMask;
+
+		(*w->screen->windowStateChangeNotify) (w);
+
+		setWindowState (w->screen->display, w->state, w->id);
+	    }
+	}
+
 	return;
+    }
 
     /* transition from minimized to shaded */
     if (w->state & CompWindowStateShadedMask)
@@ -3789,6 +4022,8 @@ showWindow (CompWindow *w)
 
     w->state &= ~CompWindowStateHiddenMask;
 
+    w->pendingMaps++;
+
     XMapWindow (w->screen->display->display, w->id);
 
     (*w->screen->windowStateChangeNotify) (w);
@@ -3804,7 +4039,16 @@ minimizeTransients (CompWindow *w,
 
     if (w->transientFor == ancestor->id ||
 	isGroupTransient (w, ancestor->clientLeader))
-	minimizeWindow (w);
+    {
+	if (!w->minimized)
+	{
+	    w->minimized = TRUE;
+
+	    forEachWindowOnScreen (w->screen, minimizeTransients, (void *) w);
+
+	    hideWindow (w);
+	}
+    }
 }
 
 void
@@ -3952,6 +4196,9 @@ unredirectWindow (CompWindow *w)
 
     w->redirected = FALSE;
     w->screen->overlayWindowCount++;
+
+    if (w->screen->overlayWindowCount > 0)
+	hideOutputWindow (w->screen);
 }
 
 void
@@ -3965,6 +4212,9 @@ redirectWindow (CompWindow *w)
 
     w->redirected = TRUE;
     w->screen->overlayWindowCount--;
+
+    if (w->screen->overlayWindowCount < 1)
+	showOutputWindow (w->screen);
 }
 
 /* Returns the current viewport for a window. If the window spans more than
@@ -4181,3 +4431,40 @@ outputDeviceForWindow (CompWindow *w)
 
     return output;
 }
+
+Bool
+onCurrentDesktop (CompWindow *w)
+{
+    if (w->desktop == 0xffffffff || w->desktop == w->screen->currentDesktop)
+	return TRUE;
+
+    return FALSE;
+}
+
+void
+setDesktopForWindow (CompWindow   *w,
+		     unsigned int desktop)
+{
+    if (desktop != 0xffffffff)
+    {
+	if (w->type & (CompWindowTypeDesktopMask | CompWindowTypeDockMask))
+	    return;
+
+	if (desktop < 0 || desktop >= w->screen->nDesktop)
+	    return;
+    }
+
+    if (desktop == w->desktop)
+	return;
+
+    w->desktop = desktop;
+
+    if (desktop == 0xffffffff || desktop == w->screen->currentDesktop)
+	showWindow (w);
+    else
+	hideWindow (w);
+
+    setWindowProp (w->screen->display, w->id,
+		   w->screen->display->winDesktopAtom,
+		   w->desktop);
+}
